import 'package:hive/hive.dart';
import 'package:live_repository/src/repository/hive_repository.dart';
import 'package:live_repository/src/repository/live_repository.dart';
import 'package:live_repository/src/repository/repository.dart';
import 'package:live_repository/src/repository/repository_options.dart';
import 'package:live_repository/src/repository/sync_repository.dart';

class RepositoryFactory {
  static bool _hiveReady = false;
  
  static final List<BaseRepository> _repositories = List<BaseRepository>();
  static final List<Repository> _liveRepositories = List<Repository>();

  /// Creates a new repository with a entity type and a key type
  /// Specify the [name] of the repository, the [keySelector] used to retrieve entities and
  /// optionally you can configure the repository using [options].
  /// Don't forget to register all the required Hive adapters before creating a repository
  static Future<BaseRepository<TEntity, TKey>> createKeyedRepository<TEntity, TKey>(String name, TKey Function(TEntity entity) keySelector, {RepositoryOptions options}) async {
    if(!_hiveReady) {
      Hive.init("data");
      _hiveReady = true;
    }

    // Duplicated repository
    if(_repositories.any((element) => element.name == name)) {
      throw Exception("A repository with name '$name' already exists.");
    }

    // Open HiveBox
    Box<TEntity> box = await Hive.openBox<TEntity>(name);

    // Create repository
    var repo = HiveRepository(name, options ?? RepositoryOptions(), box, keySelector);
    _repositories.add(repo);

    return repo; 
  }

  /// Creates a new repository with a entity type that uses a [String] as primary key
  /// Specify the [name] of the repository, the [keySelector] used to retrieve entities and
  /// optionally you can configure the repository using [options].
  /// Don't forget to register all the required Hive adapters before creating a repository
  static Future<BaseRepository<TEntity, String>> createRepository<TEntity>(String name, String Function(TEntity entity) keySelector, {RepositoryOptions options}) async {
    if(!_hiveReady) {
      Hive.init("data");
      _hiveReady = true;
    }

    // Duplicated repository
    if(_repositories.any((element) => element.name == name)) {
      throw Exception("A repository with name '$name' already exists.");
    }

    // Open HiveBox
    Box<TEntity> box = await Hive.openBox<TEntity>(name);

    // Create repository
    var repo = HiveRepository(name, options ?? RepositoryOptions(), box, keySelector);
    _repositories.add(repo);

    return repo; 
  }

  /// Finds and returns a previusly created repository that stores types of [TEntity]
  /// and uses [TKey] as primary key 
  static BaseRepository<TEntity, TKey> getKeyedRepository<TEntity, TKey>() {

    try {
      BaseRepository<TEntity, TKey> repository = _repositories.whereType<BaseRepository<TEntity, TKey>>().first;

      return repository;
    } catch (e) {
      throw Exception("Repository of type $TEntity and key $TKey not found.");
    }

  }

  /// Finds and returns a previusly created repository that stores types of [TEntity]
  /// and uses a [String] as primary key 
  static BaseRepository<TEntity, String> getRepository<TEntity>() {

    try {
      BaseRepository<TEntity, String> repository = _repositories.whereType<BaseRepository<TEntity, String>>().first;

      return repository;
    } catch (e) {
      throw Exception("Repository of type $TEntity not found.");
    }

  }

  /// Finds and returns a previusly created repository searching by its name.
  /// The repository returned is composed of [dynamic] types 
  static BaseRepository getRepositoryFromName(String name) {

    try {
      BaseRepository repository = _repositories.firstWhere((element) => element.name == name);

      return repository;
    } catch (e) {
      throw Exception("Repository with name '$name' not found.");
    }

  }

  /// Creates a new Live Repository that can be synchronized with a remote server
  static Future<Repository<TEntity, TKey>> createKeyedLiveRepository<TEntity, TKey>({
    String name,
    TKey Function(TEntity entity) keySelector,
    String Function(TEntity entity) updateStampSelector,
    LiveRepository<TEntity, TKey> liveRepository, 
    RepositoryOptions options}) async {

    if(!_hiveReady) {
      Hive.init("data");
      _hiveReady = true;
    }

    if(name == null || name.isEmpty) throw Exception("Repository name cannot be null or empty.");
    if(keySelector == null) throw Exception("Key selector cannot be null.");
    if(updateStampSelector == null) throw Exception("Update stamp selector cannot be null.");
    if(liveRepository == null) throw Exception("LiveRepository cannot be null.");

    // Open HiveBox
    Box<TEntity> box = await Hive.openBox<TEntity>(name);

    // Duplicated repository
    if(_liveRepositories.any((element) => element.name == name)) {
      throw Exception("A repository with name '$name' already exists.");
    }

    // Create repository
    Repository<TEntity, TKey> repository = Repository(
      cacheRepository: HiveRepository<TEntity, TKey>(name, options, box, keySelector),
      liveRepository: liveRepository,
      name: name,
      updateStampFunction: updateStampSelector,
      keySelector: keySelector
    );

    // Add to list
    _liveRepositories.add(repository);

    // Return the repository
    return repository;
  }

  /// Creates a new Live Repository that can be synchronized with a remote server
  /// and uses a [String] as primary key.
  static Future<Repository<TEntity, String>> createLiveRepository<TEntity>({
    String name,
    String Function(TEntity entity) keySelector,
    String Function(TEntity entity) updateStampSelector,
    LiveRepository<TEntity, String> liveRepository, 
    RepositoryOptions options}) async {

    if(!_hiveReady) {
      Hive.init("data");
      _hiveReady = true;
    }

    if(name == null || name.isEmpty) throw Exception("Repository name cannot be null or empty.");
    if(keySelector == null) throw Exception("Key selector cannot be null.");
    if(updateStampSelector == null) throw Exception("Update stamp selector cannot be null.");
    if(liveRepository == null) throw Exception("LiveRepository cannot be null.");

    // Open HiveBox
    Box<TEntity> box = await Hive.openBox<TEntity>(name);

    // Duplicated repository
    if(_liveRepositories.any((element) => element.name == name)) {
      throw Exception("A repository with name '$name' already exists.");
    }

    // Create repository
    Repository<TEntity, String> repository = Repository(
      cacheRepository: HiveRepository<TEntity, String>(name, options, box, keySelector),
      liveRepository: liveRepository,
      name: name,
      updateStampFunction: updateStampSelector,
      keySelector: keySelector
    );

    // Add to list
    _liveRepositories.add(repository);

    // Return the repository
    return repository;
  }

  /// Returns a [Repository] which stores [TEntity] types and uses a 
  /// [String] as primary key
  static Repository<TEntity, String> getLiveRepository<TEntity>() {
    try {
      Repository<TEntity, String> repository = _liveRepositories.whereType<Repository<TEntity, String>>().first;

      return repository;
    } catch (e) {
      throw Exception("Repository of type $TEntity not found.");
    }
  }

  /// Returns a [Repository] which stores [TEntity] types and uses a 
  /// [TKey] as primary key
  static Repository<TEntity, TKey> getKeyedLiveRepository<TEntity, TKey>() {
    try {
      Repository<TEntity, TKey> repository = _liveRepositories.whereType<Repository<TEntity, TKey>>().first;

      return repository;
    } catch (e) {
      throw Exception("Repository of type $TEntity not found.");
    }
  }
}