import 'package:meta/meta.dart';
import 'package:uuid/uuid.dart';

/// Base class used fetch data from a remote server.
/// Manages base CRUD operations
/// [TEntity] is the entity type to be fetched
/// [TKey] is the key type to use as primary key
abstract class LiveRepository<TEntity, TKey>{

  String _id;

  String get id => _id;

  @protected
  String Function(TEntity) updateStampSelector;

  LiveRepository() {
    _id = Uuid().v4();
  }

  /// Pulls an entity searching by its id
  Future<TEntity> fetchEntity(TKey id);

  /// Pulls many entities searching by their ids
  Future<List<TEntity>> fetchEntities(List<TKey> ids);

  /// Pulls an entity searching by a predication
  Future<TEntity> fetchEntityWhere(bool Function(TEntity) predication);

  /// Pulls many entities searching by a predication
  Future<TEntity> fetchEntitiesWhere(bool Function(TEntity) predication);

  /// Deletes an entity by its id.
  /// Returns [true] if the entity was deleted successfully, otherwise, false.
  Future<bool> deleteEntity(TKey id);

  /// Deletes many entities by their ids.
  /// Returns an [int] with the amount of entities removed, if any.
  Future<int> deleteEntities(List<TKey> ids);

  /// Stores an entity using an id
  /// Returns [true] if the entity was saved successfully, otherwise, false.
  Future<bool> putEntity(TKey key, TEntity entity);

  /// Stores many entities using their ids.
  /// Returns an [int] with the many of entities saved, if any.
  Future<int> putEntities(Map<TKey, TEntity> entities);

  /// Updates an entity, searching it by its [id]
  /// Returns [true] if the entity was updated successfully 
  Future<bool> updateEntity(TKey id, TEntity newEntity);

  /// Updates many entities, searching them by their [ids]
  /// Returns an [int] with the amount of entities updated, if any.
  Future<int> updateEntities(Map<TKey, TEntity> entities);

  /// Returns true if an entity exists 
  Future<bool> existsEntity(TKey key);

  /// Checks if an entity is up to date
  Future<bool> isUpToDate(TKey entityId, String updateStamp);

  /// Returns the amount of entities
  Future<int> countEntities();
}