import 'package:live_repository/src/exceptions/exceptions.dart';
import 'package:live_repository/src/helpers/logger.dart';
import 'package:live_repository/src/operation/base_operation.dart';
import 'package:live_repository/src/operation/operations.dart';
import 'package:live_repository/src/repository/repository_options.dart';
import 'package:meta/meta.dart';
import 'package:uuid/uuid.dart';

/// Represents a base repository
abstract class BaseRepository<TEntity, TKey> {

  String _repoName, _repoId;
  RepositoryOptions _options;
  bool _empty;
  List<BaseOperation> _pendingOperations;

  @protected
  TKey Function(TEntity entity) keySelector;

  /// The name of the repository
  String get name => _repoName;

  /// The unique id for this repository
  String get id => _repoId;

  /// Returns true if the repository is empty, false if not.
  bool get isEmpty => _empty;

  /// Returns true if the repository is not empty, true if it.
  bool get isNotEmpty => !_empty;

  /// Contains all the entities in the repository
  List<TEntity> get entities;

  BaseRepository({String repositoryName, RepositoryOptions options, this.keySelector}) {

    if(repositoryName == null || repositoryName.isEmpty) {
      throw Exception("Repository name is required.");
    }

    if(options == null) {
      _options = RepositoryOptions();
    }

    if(keySelector == null) {
      throw Exception("Key selector must be specified.");
    }

    _repoName = repositoryName;
    _options = options;
    _repoId = Uuid().v4();
    _pendingOperations = List<BaseOperation>();

    Logger.information("Registering new repository with name '$name' and id '$_repoId'");
  }

  @protected
  StoreOperation<TEntity> createStoreOperation(TKey key, TEntity value);

  @protected
  BatchStoreOperation<TEntity> createBatchStoreOperation(Map<TKey, TEntity> entities);

  @protected
  DeleteOperation<TEntity> createDeleteOperation(TKey key, TEntity value);

  @protected
  UpdateOperation<TEntity> createUpdateOperation(TKey key, TEntity updatedValue);

  @protected
  @visibleForTesting
  bool existsWrapper(TKey key);

  @protected
  @visibleForTesting
  TEntity findWrapper(TKey key);

  /// Returns an entity searching by its id.
  /// If the entity does not exist, it will throw an [EntityNotFoundException] if it was configured while creating the repository,
  /// otherwise, it will return null.
  TEntity getEntityFromId(TKey key) {
    try {
      TEntity entity = findWrapper(key);

      if(entity == null) {
        if(_options.throwExceptionOnNotFound) {
          throw EntityNotFoundException("$key");
        }

        return null;
      }

      return entity;

    } catch (e) {
      if(_options.throwExceptionOnNotFound) {
        throw e;
      }

      return null;
    }
  }

  /// Stores an entity.
  /// If the entity already exists, it will throw an [DuplicatedEntityException] if it was configured while creating the repository,
  /// otherwise, it will be ignored.
  void store(TEntity entity) {
    _pendingOperations.add(createStoreOperation(keySelector(entity), entity));
  }

  /// Stores many entities at once.
  /// If any entity already exists, it will throw an [DuplicatedEntityException] if it was configured while creating the repository,
  /// otherwise, it will be ignored.
  void batchStore(List<TEntity> entities) {
    _pendingOperations.add(createBatchStoreOperation(Map.fromIterable(entities, key: (v) => v[0], value: (v) => v[1])));
  }

  /// Deletes an entity.
  /// If the entity does not exist, it will throw an [EntityNotFoundException] if it was configured while creating the repository,
  /// otherwise, it will be ignored.
  void delete(TEntity entity) {
    _pendingOperations.add(createDeleteOperation(keySelector(entity), entity));
  }

  /// Deletes an entity searching it using its id.
  /// If the entity does not exist, it will throw an [EntityNotFoundException] if it was configured while creating the repository,
  /// otherwise, it will be ignored.
  void deleteFromId(TKey key) {
    TEntity entity = getEntityFromId(key);

    if(entity == null) {
      if(_options.throwExceptionOnNotFound) {
        throw EntityNotFoundException("$key");
      }
      return;
    }

    delete(entity);
  }

  /// Updates an entity.
  /// If the entity does not exist, it will throw an [EntityNotFoundException] if it was configured while creating the repository,
  /// otherwise, it will be ignored.
  void update(TEntity entity) {
    _pendingOperations.add(createUpdateOperation(keySelector(entity), entity));
  }

  /// Returns true if an entity exists in a repository or false if not.
  /// Takes as parameter the id of the entity to check.
  bool exists(TKey key) {
    return existsWrapper(key);
  }

  /// Returns true if an entity exists in a repository or false if not.
  /// Takes as parameter the entity to check.
  bool existsEntity(TEntity entity) {
    return exists(keySelector(entity));
  }

  /// Returns the id of an entity
  TKey getEntityId(TEntity entity) {
    return keySelector(entity);
  }

  /// Saves all the changes made in the repository.
  Future saveChangesAsync() async {

    Logger.information("Saving ${_pendingOperations.length} changes...");

    List<BaseOperation> errors = List<BaseOperation>();

    for (BaseOperation operation in _pendingOperations) {
      bool result = await operation.executeAsync();

      if(!result) {
        errors.add(operation);
      }
    }

    if(errors.length > 0) {
      Logger.warning("${errors.length} changes could not be saved.");
      errors.clear();
    }

    _pendingOperations.clear();

    Logger.information("Changes saved.");
  }

}