import 'package:live_repository/src/helpers/logger.dart';
import 'package:live_repository/src/repository/live_repository.dart';
import 'package:live_repository/src/repository/repository.dart';

/// Contains a cached repository and a sync repository
/// [TEntity] is the entity type to be stored
/// [TKey] is the key type to use as primary key
class Repository<TEntity, TKey> {

  String _name;
  BaseRepository<TEntity, TKey> _cacheRepository;
  LiveRepository<TEntity, TKey> _syncRepository;
  TKey Function(TEntity entity) _keySelector;

  String Function(TEntity) updateStampFunction;
  String get name => _name;

  Repository({
    String name,
    BaseRepository<TEntity, TKey> cacheRepository, 
    LiveRepository<TEntity, TKey> liveRepository,
    this.updateStampFunction,
    TKey Function(TEntity) keySelector}) {

    if(cacheRepository == null) {
      throw Exception("Cache repository cannot be null.");
    }

    if(liveRepository == null) {
      throw Exception("Sync repository cannot be null.");
    }
    
    _name = name;
    _cacheRepository = cacheRepository;
    _syncRepository = liveRepository;
    _keySelector = keySelector;
  }

  /// Returns an entity searching by its id
  Future<TEntity> getEntity(TKey id) async {

    // Get entity from cache repository, if exists
    if(_cacheRepository.exists(id)) {
      TEntity entity = _cacheRepository.getEntityFromId(id);

      // Check if its up to date 
      bool upToDate = await _syncRepository.isUpToDate(id, updateStampFunction(entity));

      // If the entity is up to date, return it
      if(upToDate) {
        return entity;

      // Otherwise, sync
      } else {
        TEntity entity = await _syncRepository.fetchEntity(id);

        // If the entity does not exists in live repository, return null and delete from cache
        if(entity == null) {
          _cacheRepository.deleteFromId(id);
          await _cacheRepository.saveChangesAsync();

          return null;
        }
      }

    // Try to get from server
    } else {
      
      // If the entity exists, fetch it and save
      if(await _syncRepository.existsEntity(id)) {
        
        TEntity entity = await _syncRepository.fetchEntity(id);
        _cacheRepository.store(entity);
        await _cacheRepository.saveChangesAsync();

        return entity;
      } else {
        // Otherwise, return null
        return null;
      }
    }

    return null;
  }

  /// Stores an entity
  Future storeEntity(TEntity entity) async {

    // Return false if the entity already exists
    if(_cacheRepository.existsEntity(entity) || await _syncRepository.existsEntity(_keySelector(entity))) {
      return false;
    }

    // Store entity locally and remotelly
    _cacheRepository.store(entity);
    await _cacheRepository.saveChangesAsync();

    bool result = await _syncRepository.putEntity(_keySelector(entity), entity);

    return result;
  }

  /// Stores many entities at once
  Future storeEntities(List<TEntity> entities) async {
    for(TEntity entity in entities) {
      if(_cacheRepository.existsEntity(entity)) {
        continue;
      }

      _cacheRepository.store(entity);

      // Store remotelly
      if(await _syncRepository.existsEntity(_keySelector(entity))) {
        continue;
      }

      await _syncRepository.putEntity(_keySelector(entity), entity);
    }

    // Save changes locally
    await _cacheRepository.saveChangesAsync();
  }

  /// Deletes an entity
  Future deleteEntity(TEntity entity) async {
    // Return false if the entity already exists
    if(_cacheRepository.existsEntity(entity) || await _syncRepository.existsEntity(_keySelector(entity))) {
      return false;
    }

    // Delete
    _cacheRepository.delete(entity);
    await _cacheRepository.saveChangesAsync();
    return await _syncRepository.deleteEntity(_keySelector(entity));
  }

  /// Deletes many entities at once
  Future deleteEntities(List<TEntity> entities) async {
    for(TEntity entity in entities) {

      // Delete locally
      _cacheRepository.delete(entity);

      // Delete remotelly
      if(await _syncRepository.existsEntity(_keySelector(entity))) {
        await _syncRepository.deleteEntity(_keySelector(entity));
      }
    }

    // Save changes locally
    await _cacheRepository.saveChangesAsync();
  }

  /// Updates an entity
  Future updateEntity(TEntity entity) async {
    TKey entityId = _keySelector(entity);

    // Check if the entity exists in the live repository
    if(await _syncRepository.existsEntity(entityId)) {
      
      bool result = await _syncRepository.updateEntity(entityId, entity);
      
      // Update locally
      if(result) {
        _cacheRepository.update(entity);
        await _cacheRepository.saveChangesAsync();

        return true;
      }

      return false;

    } else {
      // Return false if not
      return false;
    }
  }

  /// Synchronizes both repositories
  Future syncData() async {

    Logger.information("Syncing data...");

    int localEntities = _cacheRepository.entities.length;
    int remoteEntities = await _syncRepository.countEntities();

    // Download missing entities from remote
    if(remoteEntities > localEntities) {
      Logger.information("Download missing...");
      await _downloadMissing();
    }
    // Upload local
    else if(remoteEntities < localEntities) {
      Logger.information("Uploading local entities...");
      await _uploadLocal();
    }

    // Update outdated entities
    else {
      Logger.information("Updating local entities...");
      await _updateEntities();
    }

  }

  /// Returns the Cache Repository instance
  BaseRepository<TEntity, TKey> getCacheRepository() => _cacheRepository;

  /// Returns the Live Repository instance
  LiveRepository<TEntity, TKey> getLiveRepository() => _syncRepository;

  /// #region Helper Methods

  Future _updateEntities() async {
    for(TEntity entity in _cacheRepository.entities) {

      // Ignore if up to date
      if(!await _syncRepository.isUpToDate(_keySelector(entity), updateStampFunction(entity))) {
        continue;
      }

      // Update locally
      TEntity remoteEntity = await _syncRepository.fetchEntity(_keySelector(entity));

      // Delete locally
      if(remoteEntity == null) {
        _cacheRepository.delete(entity);
      }

      // Store
      _cacheRepository.update(entity);
    }

    // Save changes
    await _cacheRepository.saveChangesAsync();
    Logger.information("Local entities updated...");
  }

  Future _uploadLocal() async {
    for(TEntity entity in _cacheRepository.entities) {

      // Ignore if already exists
      if(await _syncRepository.existsEntity(_keySelector(entity))) {
        continue;
      }

      // Store remotelly
      await _syncRepository.putEntity(_keySelector(entity), entity);
    }

    Logger.information("Local entities uploaded...");
  }

  Future _downloadMissing() async {
    // THIS METHOD IS OMITTED BECAUSE OF AMOUNT OF DATA.
    // IF THOUSANDS OF ENTITIES ARE IN THE SERVER, MAYBE NOT ALL THE DATA WANT TO BE DOWNLOADED 
  }

  /// #endregion

}